#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>
#include <unistd.h>

static char SccsId[] = "@(#)super.c	1.12\t4/13/93";

#ifndef SUPERFILE
#define SUPERFILE "/usr/local/lib/super.tab"
#endif

/*
 *	Copyright (c) 1993 by California Institute of Technology.
 *	Written by William Deich.  Not derived from licensed software.

 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of this
 *	   software, no matter how awful, even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 */

/*
 * Super allows users to execute other programs (particularly
 * scripts) as root, without unduly compromising security.
 * 
 * Use:
 * 
 *     $0 commandname args...
 * 
 * If the commandname is "-h" or "-?", or missing, super prints its current
 * list of allowed commands, but nothing executed.
 * 
 * The super.tab file names each command that super will execute, and
 * says who can use it. It contains lines like:
 * 
 *     commandname   fullpathname    valid-user/group ed-type patterns
 * 
 * See sample.tab for how to specify user & group patterns.
 * If a user is allowed to execute a given <commandname>, the <fullpathname>
 * is exec'd, with <commandname> as argv[0].
 * 
 * For security, the environment variables are discarded, save for TERM,
 * LINES, and COLUMNS.  If TERM contains any characters other than
 * [a-z][A-Z][0-9]_, it is discarded.  If LINES or COLUMNS contains any
 * characters other than [0-9], they are discarded.  To these are added
 * reasonable values for IFS, PATH, USER and HOME (USER and HOME are set
 * to the username and login directory, respectively, of the person who
 * who runs super).  LOGNAME is set to the same as USER.  SUPERCMD is set
 * to the <command>.  All descriptors excepting 0,1,2 are closed.  Signals
 * are all reset to have default handling.
 */


static char *SAFE_IFS = "IFS= \t\n";
static char *SAFE_PATH = "PATH=/bin:/usr/bin:/usr/ucb";

char *prog;	/* This program */
char *user;	/* who's invoking prog */

main(argc, argv)
int argc;
char **argv;
{
    char *s, *path, *approve();
    char **buttonup(), **envp;
    void exit();

    s = strchr(argv[0], '/');
    prog = (s && *(s+1)) ? s+1 : argv[0];

    if ((path = approve(SUPERFILE, argv[1])) == NULL)
	(void) exit(1);
    else if (*path == '\0')
	(void) exit(0);

    /* Button up for security, and get a modified environment */
    envp = buttonup(argv[1]);

    if (execve(path, &argv[1], envp) == -1) {
	(void) fprintf(stderr, "%s: Couldn't exec %s (%s): ",
					prog, argv[1], path);
	perror("");
	(void) exit(1);
    }
    return 0;
}

char **
buttonup(cmd)
char *cmd;	/* name of command being started */
{
    /* Closes all descriptors save 0,1,2.
     * Resets all signal-handling to SIG_DFL.
     * Discards all env. variables save for TERM, LINES, and COLUMNS.
     * Don't allow TERM to have any but alpha characters and underscore.
     * Don't allow LINES, COLUMNS to have anything but digits.
     * To these are added reasonable values for IFS, PATH, USER, and HOME.
     * LOGNAME is set to the same as USER, and SUPERCMD is set to cmd.
     * Returned:
     *	a pointer to the modified environment list.
     */
    int i, fd, n;
    char *Getenv();
    int getlogdir();
    int checkenv();
    static char *env[10];
    static char User[100];		/* USER */
    static char Logname[100];		/* LOGNAME (alias for USER) */
    static char Home[MAXPATHLEN+5];	/* HOME */
    static char Cmd[1200];		/* SUPERCMD */
    void (*signal())();

#ifdef MAXFD
    n = MAXFD ;
#else
    n = getdtablesize(); 
#endif
    for (fd=3; fd <= n; fd++)
	(void) close(fd);
    
    for (i=0; i<NSIG; i++)
       (void) signal(i, SIG_DFL);

    (void) sprintf(User, "USER=%s", user);
    (void) sprintf(Logname, "LOGNAME=%s", user);
    (void) sprintf(Cmd, "SUPERCMD=%s", cmd);
    (void) strcpy(Home, "HOME=");
    (void) getlogdir(user, Home+5);
    i = 0;
    env[i] = Getenv("TERM");
    if (env[i] && checkenv("TERM", env[i]+5, "^[a-zA-Z0-9]*$")) i++;
    env[i] = Getenv("LINES");
    if (env[i] && checkenv("LINES", env[i]+6, "^[0-9]*$")) i++;
    env[i] = Getenv("COLUMNS");
    if (env[i] && checkenv("COLUMNS", env[i]+8, "^[0-9]*$")) i++;
    env[i++] = SAFE_IFS;
    env[i++] = SAFE_PATH;
    env[i++] = User;
    env[i++] = Logname;
    env[i++] = Cmd;
    env[i] = (char *) NULL;

    return &env[0];
}

char *
Getenv(s)
char *s;
{
    /* Like getenv(), but returns ptr to the <name> in "name=xxxx",
     * not just the xxxx.
     */
    char **envp; 
    int l;
    extern char **environ;

    if (!s)
	return (char *) NULL;
    l = strlen(s);
    for (envp=environ; *envp ; envp++)
	if (strncmp(*envp, s, l) == 0  &&  *(*envp+l) == '=')
	    return *envp;
    return (char *) NULL;
}

/* Returns 1 if pat matched; 0 otherwise.  */
checkenv(name, value, pat)
char *name;	/* variable name to check (e.g. "TERM") */
char *value;	/* contents of variable (e.g. "vt100") */
char *pat;	/* pattern that value must match */
{
    char *re_comp();
    int re_exec();

    if (!value)
	return 0;
    
    if (re_comp(pat)) {
	(void) fprintf(stderr,
	    "%s: Warning: couldn't compile pattern `%s'\n", prog, pat);
	return 0;
    }

    if (re_exec(value) != 1) {
	(void) fprintf(stderr,
		"%s: %s (%s) doesn't match %s.\n", prog, name, value, pat);
	return 0;
    }
    return 1;
}
